#ifndef DISPLAYH
#define DISPLAYH

#include "mesh.h"
#include "terrain.h"
#include "console.h"
#include "image.h"
#include "menu.h"
#include "skeleton.h"
#include "object.h"

#include <list>
#include <map>
#include <string>

class Display
{
	bool displayNormals;
	int windowWidth;
	int windowHeight;
	std::list<unsigned int> textures;
public:
	Display();
	~Display();
	void perspective(float fov, float aspect, float zNear, float zFar);
	void ortho(float fleft, float fright, float fbottom, float ftop, float fnear, float ffar);
	void depthTest(bool depthTest = true);
	void resize(int x, int y);
	void display(Mesh &mesh);
	void display(Terrain &terrain);
	int width() { return windowWidth; }
	int height() { return windowHeight; }

	void display(Console &console);
	void display(MenuData &menu);
	void display(SkeletonNode &skeleton, std::map<std::string, Mesh> &meshTable);
	void display(Object &object, std::map<std::string, Mesh> &meshTable);
	
	unsigned int load(Image &texture);
	void select(unsigned int texture);
	void release(unsigned int texture);

	void normals(bool b);
	void wireframe(bool b);
	// I want to see Cel-Shading here some day as a display option.
	
	void axes();
	void rectangle(float left, float right, float bottom, float top);
	void text(const char *message, float left, float bottom, float charWidth, float charHeight);
	
	void lightPosition(int lightNumber, const vertex &position);
	void lightDiffuse(int lightNumber, const vertex &diffuse);
	void lightEnable(int lightNumber, bool lightEnabled);
	void lighting(bool lightEnabled);

	void fog(bool enabled = true);
	void fogColor(float red, float green, float blue, float density);
	void fogCutoff(float start, float end);

	void clear();
	void camera(const vertex &position, const vertex &rotation);
	
	void color(float red, float green, float blue);
};

#endif
