#include "gamecore.h"

#include <iostream>
#include <math.h>
#include <sstream>
#include <stack>
#include <string>
#include <stdlib.h>

#include "explosive.h"
#include "explosion.h"
#include "image.h"

#include "window.h" // gives us information about the keys that are passed to keypress...

GameCore::GameCore()
: terrain(64, 64, 200.0f, 200.0f, 200.0f), consoleActive(false), frames(0), fps(0), missile(ExplosiveWeapon::Missile)
{
	std::cout << "GameCore::GameCore()" << std::endl;

	// Camera Position
	rotation.x = rotation.y = rotation.z = 0;
	position.x = 50.0f;
	position.y = 25.0f;
	position.z = 100.0f;

	// Initialize the table storing key states to false (no keys pressed)
	for (int c = 0; c < 256; keyTable[c++] = false);

	cameraVelocity.x = cameraVelocity.y = cameraVelocity.z = 0.0f;

	console.print("Tank Game");

	aimPhi = aimTheta = 0.0f;
	velocity = 10.0f; // Initial projectile velocity
	sAimPhi = sAimTheta = "0";
	sVelocity = "10";
	sHealth = "100";

	displayMessage = true;
	message = "Press Any Key to Begin";

	// Initialize some run-time variables
	variableTable["normals"] = false;
	variableTable["keydump"] = false;
	variableTable["wireframe"] = false;
	variableTable["lighting"] = true;
	variableTable["menu"] = false;
	variableTable["fps"] = false;
	splashScreen = true;
	activeMenu = 0;

	// Set lighting parameters
	light.x = light.y = light.z = 200.0f; // Light position
	vertex white = {1.0f, 1.0f, 1.0f};
	disp.lighting(true);
	disp.lightEnable(0, true);
	disp.lightDiffuse(0, white);

	// Try some spiffy fog effects (disabled, because I don't know where this has its place)
	//disp.fog(true);
	//disp.fogColor(0.0f, 0.0f, 0.0f, 0.005f);
	//disp.fogCutoff(200.0f, 260.0f);

	// Load images
	texMissile = imageID = loadImage("data/giraffe.png");
	font = loadImage("data/font.png");
	texExplosion = loadImage("data/fire.png");
	texSelect = loadImage("data/selection.png");
	texTerrain[0][0] = loadImage("data/terrain/00.png");
	texTerrain[0][1] = loadImage("data/terrain/01.png");
	texTerrain[0][2] = loadImage("data/terrain/02.png");
	texTerrain[0][3] = loadImage("data/terrain/03.png");
	texTerrain[0][4] = loadImage("data/terrain/04.png");
	texTerrain[1][0] = loadImage("data/terrain/10.png");
	texTerrain[1][1] = loadImage("data/terrain/11.png");
	texTerrain[1][2] = loadImage("data/terrain/12.png");
	texTerrain[1][3] = loadImage("data/terrain/13.png");
	texTerrain[1][4] = loadImage("data/terrain/14.png");
	texTerrain[2][0] = loadImage("data/terrain/20.png");
	texTerrain[2][1] = texTerrain[2][2] = texTerrain[2][3] = texTerrain[2][4] = texTerrain[2][0];
	texBackground = loadImage("data/background.png");
	texSplashScreen = loadImage("data/splashscreen.png");

	std::list<Terrain::TextureInfo> textureInfo;
	Terrain::TextureInfo t;
	t.height = 0.4f;
	t.textures[0] = texTerrain[0][0];
	t.textures[1] = texTerrain[0][1];
	t.textures[2] = texTerrain[0][2];
	t.textures[3] = texTerrain[0][3];
	t.textures[4] = texTerrain[0][4];
	textureInfo.push_back(t);
	t.height = 0.55f;
	t.textures[0] = texTerrain[1][0];
	t.textures[1] = texTerrain[1][1];
	t.textures[2] = texTerrain[1][2];
	t.textures[3] = texTerrain[1][3];
	t.textures[4] = texTerrain[1][4];
	textureInfo.push_back(t);
	t.height = 1.0;
	t.textures[0] = texTerrain[2][0];
	t.textures[1] = texTerrain[2][1];
	t.textures[2] = texTerrain[2][2];
	t.textures[3] = texTerrain[2][3];
	t.textures[4] = texTerrain[2][4];
	textureInfo.push_back(t);

	terrain.setTextureInfo(textureInfo);
	terrain.generate(300, 0.25f, 0.75);

	std::cout << "Loading skeleton: " << load("data/tank.skel", skeleton) << std::endl;
	std::cout << "Loading skeleton: " << load("data/sphere.skel", sphere) << std::endl;
	std::cout << "Loading skeleton: " << load("data/explosion.skel", explosion) << std::endl;
	std::cout << "Loading skeleton: " << load("data/background.skel", skelBackground) << std::endl;

	Weapon *w = new ExplosiveWeapon(ExplosiveWeapon::BabyMissile);
	w->load(meshMap, disp);
	weapons.push_back(w);
	w = new ExplosiveWeapon(ExplosiveWeapon::Missile);
	w->load(meshMap, disp);
	weapons.push_back(w);
	w = new ExplosiveWeapon(ExplosiveWeapon::BabyNuke);
	w->load(meshMap, disp);
	weapons.push_back(w);
	w = new ExplosiveWeapon(ExplosiveWeapon::Nuke);
	w->load(meshMap, disp);
	weapons.push_back(w);
	missile.load(meshMap, disp);
	w = new MIRV;
	w->load(meshMap, disp);
	weapons.push_back(w);
	w = new DeathsHead;
	w->load(meshMap, disp);
	weapons.push_back(w);
	selectedWeapon = weapons.begin();

	std::cout << "Loading menu: " << std::flush << load("data/menu.xml", menu) << std::endl;

	message = "";
	displayMessage = true;
	tank = tanks.end();

	// Now that we have the skeleton loaded, we need to load all of the meshes that it calls upon.
	std::stack<SkeletonNode *> nodes;
	nodes.push(&skeleton);
	nodes.push(&sphere);
	nodes.push(&explosion);
	nodes.push(&skelBackground);

	while (nodes.size())
	{
		SkeletonNode *node = nodes.top();
		nodes.pop();

		for (std::vector<SkeletonNode>::iterator i = node->child.begin(); i != node->child.end(); nodes.push(&*i++)) {}

		if (node->mesh.size()) load(("data/" + node->mesh).c_str(), meshMap[node->mesh]);
	}

	std::cout << "end GameCore::GameCore()" << std::endl;
}

GameCore::~GameCore()
{
	std::cout << "~GameCore()" << std::endl; // This never seems to be called.  Look into an alternative way to exit

	std::list<Object *>::iterator i = objects.begin();

	while (i != objects.end())
	{
		delete *i;
		i = objects.erase(i);
	}

	std::list<Weapon *>::iterator w = weapons.begin();

	while (w != weapons.end())
	{
		delete *w;
		w = weapons.erase(w);
	}
}

void GameCore::keypress(unsigned char key, bool down, int modifiers)
{
	if (variableTable["keydump"])
		std::cout << (down?"KEYDOWN: ":"KEYUP: ") << static_cast<int>(key) << " = \'" << key << '\'' << std::endl;
	
	if (splashScreen && down && !variableTable["menu"])
	{
		variableTable["menu"] = true;
		activeMenu = 0;
		return;
	}
	if (key == '`' && down)
	{
		consoleActive = !consoleActive;
		if (console.current().size())
			console.current("");
	}
	else if (key == TWK_ESCAPE && down)
	{
		displayMessage = false;
		variableTable["menu"] = !variableTable["menu"];
		activeMenu = 0;
		if (0 < menu.size())
			menu[0].selected = 0;
	}
	else if (consoleActive && down)
	{
		switch (key)
		{
		case TWK_ENTER:
			if (console.current().size())
			{
				std::string command = console.current();
				console.finalizeCurrent();
				processCommandList(command);
			}
			break;
		case TWK_BACKSPACE: // Backspace
			console.current(console.current().substr(0, console.current().size() - 1));
			break;
		default:
			if (key >= 32 && key < 128) // printable characters only...
				console.current(console.current() + static_cast<char>(key));
		}
	}
	else if (variableTable["menu"] && down && activeMenu < menu.size())
	{
		MenuData &active = menu[activeMenu];
		switch (key)
		{
		case '\r':
		case '\n':
		case ' ':
			if (active.selected >= 0 && active.selected < active.data.size())
				processCommandList(active.data[active.selected].execute);
			break;
		case TWK_UP:
			if (active.selected > 0)
				active.selected--;
			break;
		case TWK_DOWN:
			if (active.selected < active.data.size() - 1)
				active.selected++;
			break;
		case TWK_LEFT:
			if (active.data[active.selected].choices.size() && active.data[active.selected].selected > 0)
				active.data[active.selected].selected--;
			break;
		case TWK_RIGHT:
			if (active.data[active.selected].choices.size() && static_cast<unsigned int>(active.data[active.selected].selected) < active.data[active.selected].choices.size() - 1)
				active.data[active.selected].selected++;
			break;
		}
	}
	else
	{
		keyTable[key] = down;
		if (key == '\t' && down)
		{
			selectedWeapon++;
			if (selectedWeapon == weapons.end())
				selectedWeapon = weapons.begin();
		}
		if (key == ' ' && down && tank != tanks.end()) // FIRE!!!
		{
			std::map<std::string, SkeletonNode *>::iterator turret = jointTable.find("turret");
			std::map<std::string, SkeletonNode *>::iterator barrel = jointTable.find("barrel");
			if (turret != jointTable.end() && barrel != jointTable.end())
			{
				vertex newPosition;// = tank.position;
				vertex newVelocity;
				newPosition.x = newPosition.y = newPosition.z = 0.0f;
				transformVertex((*tank)->skeleton, newPosition, "tip");
				// Dan Harris gave me some help with his Calculus notes here.
				newPosition = newPosition + (*tank)->position;
				float r = velocity;
				float theta = (*turret).second->rotation.y * 3.14159f / 180.0f;
				float phi = (*barrel).second->rotation.x * 3.14159f / 180.0f;
				newVelocity.x = r * cos(phi) * sin(theta);
				newVelocity.y = -r * sin(phi);
				newVelocity.z = r * cos(phi) * cos(theta);
				(*selectedWeapon)->fire(objects, newPosition, newVelocity);
			}

			tank++;
			if (tank == tanks.end())
				tank = tanks.begin();
			getJoints();
		}
	}
}

void GameCore::mouseMotion(bool down, float x, float y, float xrel, float yrel)
{
	float step = 90.0f;

	if (down)
	{
		rotation.x -= yrel * step;
		rotation.y -= xrel * step;
		if (rotation.x > 360.0f) rotation.x -= 360.0f;
		if (rotation.y > 360.0f) rotation.y -= 360.0f;
	}
}

void GameCore::mouseButton(int button, bool down, float x, float y)
{
	// Maybe there should be mouse input for the menu.  That'd be nice.
}

void GameCore::idle()
{
	float elapsed = idleClock.elapsed();
	float step = elapsed * 4.0f;
	double theta = rotation.y * 3.14159265358979323846 / 180.0;
	float zstep = cos(theta) * step;
	float xstep = sin(theta) * step;
	float aimStep = elapsed * 90.0f;
	
	physics.advanceState(Physics::Environment(objects, Physics::Wall(0.0f, 200.0f, 0.0f, 200.0f, 0.0f, 200.0f, Physics::Wall::bounce), terrain), elapsed);

	// This is a thin link to make sure that the tanks list is maintained like the objects list.  If a tank is deleted from the objects, it has to immediately be deleted from the tanks.
	std::list<TankObject *>::iterator t = tanks.begin();
	while (t != tanks.end())
	{
		if ((*t)->health() <= 0)
		{
			console.print(((*t)->name() + " died").c_str());

			// Delete the tank from the objects list.
			std::list<Object *>::iterator o = objects.begin();
			while (o != objects.end())
				if (*o == *t)
					o = objects.erase(o);
				else
					o++;
					
			t = tanks.erase(t);
			tank = tanks.begin();
			getJoints();
		}
		else
			t++;
	}

	if (tanks.size() == 0 && !displayMessage)
	{
		displayMessage = true;
		message = "Draw Game";
	}
	if (tanks.size() == 1 && tank != tanks.end() && !displayMessage)
	{
		displayMessage = true;
		message = (*tank)->name() + " wins!";
	}

	idleClock.update();

	if (keyTable[TWK_LEFT])
	{
		cameraVelocity.z += -xstep;
		cameraVelocity.x += -zstep;
	}
	if (keyTable[TWK_RIGHT])
	{
		cameraVelocity.z += xstep;
		cameraVelocity.x += zstep;
	}
	if (keyTable[TWK_UP])
	{
		cameraVelocity.z += -zstep;
		cameraVelocity.x += xstep;
	}
	if (keyTable[TWK_DOWN])
	{
		cameraVelocity.z += zstep;
		cameraVelocity.x += -xstep;
	}
	
	if (keyTable[TWK_PAGEUP])
		cameraVelocity.y += step; // PAGE UP
	if (keyTable[TWK_PAGEDOWN])
		cameraVelocity.y += -step; // PAGE DOWN

	// Move the camera and slow down the camera's movement a bit.
	position += cameraVelocity; // I really don't know how to get these constants right.
	cameraVelocity = cameraVelocity - (4.0f * elapsed * cameraVelocity);

	float bounds[4] = {0.0f, 0.0f, 200.0f, 200.0f};
	if (position.x < bounds[0]) position.x = bounds[0];
	if (position.z < bounds[1]) position.z = bounds[1];
	if (position.x > bounds[2]) position.x = bounds[2];
	if (position.z > bounds[3]) position.z = bounds[3];

	int x = static_cast<int>(position.x / terrain.scale().x * terrain.width());
	int y = static_cast<int>(position.z / terrain.scale().z * terrain.height());
	if (x < 0) x = 0; if (y < 0) y = 0;
	unsigned int ux = static_cast<unsigned int>(x);
	unsigned int uy = static_cast<unsigned int>(y);
	if (ux >= terrain.width()) ux = terrain.width() - 1;
	if (uy >= terrain.height()) uy = terrain.height() - 1;
	float bound2 = terrain.maxHeight(ux, uy) + 2.0f;
	if (position.y < bound2) position.y = bound2;
	if (position.y > 200.0f) position.y = 200.0f;
	
	step = 0.1f;
	if (keyTable['j'])
		light.x -= step;
	if (keyTable['l'])
		light.x += step;
	if (keyTable['i'])
		light.z -= step;
	if (keyTable['k'])
		light.z += step;
	if (keyTable['u'])
		light.y += step;
	if (keyTable['o'])
		light.y -= step;
	if (keyTable['a'])
	{
		SkeletonNode *node = jointTable["turret"];
		if (node)
		{
			node->rotation.y += aimStep;
			if (node->rotation.y > 360.0f) node->rotation.y -= 360.0f;
			aimTheta = node->rotation.y;
			std::ostringstream ss;
			ss << int(node->rotation.y);
			sAimTheta = ss.str();
		}
	}
	if (keyTable['d'])
	{
		SkeletonNode *node = jointTable["turret"];
		if (node)
		{
			node->rotation.y -= aimStep;
			if (node->rotation.y < 0.0f) node->rotation.y += 360.0f;
			aimTheta = node->rotation.y;
			std::ostringstream ss;
			ss << int(node->rotation.y);
			sAimTheta = ss.str();
		}
	}
	if (keyTable['w'])
	{
		SkeletonNode *node = jointTable["barrel"];
		if (node)
		{
			node->rotation.x -= aimStep;
			if (node->rotation.x < -90.0f) node->rotation.x = -90.0f;
			aimPhi = node->rotation.x;
			std::ostringstream ss;
			ss << -int(node->rotation.x);
			sAimPhi = ss.str();
		}
	}
	if (keyTable['s'])
	{
		SkeletonNode *node = jointTable["barrel"];
		if (node)
		{
			node->rotation.x += aimStep;
			if (node->rotation.x > 0.0f) node->rotation.x = 0.0f;
			aimPhi = node->rotation.x;
			std::ostringstream ss;
			ss << -int(node->rotation.x);
			sAimPhi = ss.str();
		}
	}
	if (keyTable['r']) velocity += aimStep;
	if (keyTable['f']) velocity -= aimStep;
	if (velocity > 100.0f) velocity = 100.0f;
	if (velocity < 0.0f) velocity = 0.0f;
	if (keyTable['r'] || keyTable['f'])
	{
		std::ostringstream ss;
		ss << int(velocity);
		sVelocity = ss.str();
	}

	if (tank != tanks.end())
	{
		std::ostringstream ss;
		ss << int((*tank)->health());
		sHealth = ss.str();
	}
}

void GameCore::display()
{
	static std::string sfps;

	// Update FPS counter
	if (timer.elapsed() > 5.0f)
	{
		fps = frames / timer.elapsed();
		timer.update();
		frames = 0;
		int t = static_cast<int>(fps);
		sfps.resize(0);
		while (t)
		{
			sfps = static_cast<char>(t % 10 + '0') + sfps;
			t /= 10;
		}
	}

	frames++;

	disp.clear();

	disp.perspective(90.0f, float(disp.width()) / float(disp.height()), 0.1f, 300.0f);
	
	// This is essentially useless now.
	disp.normals(variableTable["normals"] != 0);

	disp.camera(position, rotation);
	disp.lightPosition(0, light);

	disp.wireframe(variableTable["wireframe"] != 0);

	disp.color(1.0f, 1.0f, 1.0f);
	disp.lighting(false);
	//disp.fog(false);
	disp.select(texBackground);
	disp.display(skelBackground, meshMap);
	//disp.fog(true);
	
	// Enable lighting if chosen
	disp.lighting(variableTable["lighting"] != 0);

	// Display the terrain
	disp.display(terrain);

	// Display all of the objects.
	for (std::list<Object *>::iterator i = objects.begin(); i != objects.end(); i++)
		if ((*i)->active)
			disp.display(**i, meshMap);

	disp.color(1.0f, 1.0f, 1.0f);
	disp.lighting(false);
	disp.wireframe(false);

	// Everything from here on out uses a simpler 2D coordinate system for absolute positioning in the window.
	disp.ortho(0, disp.width(), 0, disp.height(), 0, 1.0f);
	disp.depthTest(false);

	float left = 0.0f;
	float bottom = disp.height() - 32;
	float top = bottom + 32;
	for (std::list<Weapon *>::iterator w = weapons.begin(); w != weapons.end(); w++)
	{
		disp.select((*w)->icon());
		disp.rectangle(left, left + 32, bottom, top);
		if (*w == *selectedWeapon)
		{
			disp.select(texSelect);
			disp.rectangle(left, left + 32, bottom, top);
		}
		left += 32;
	}
	disp.select(font);
	disp.color(1.0f, 1.0f, 1.0f);
	if (tank != tanks.end())
	{ // Display Tank information heads-up display
		disp.text((*tank)->name().c_str(), 0, 0, 16, 16);
		disp.text(("Angle: " + sAimTheta).c_str(), 0, 16, 16, 16);
		disp.text(("Elevation: " + sAimPhi).c_str(), 0, 32, 16, 16);
		disp.text(("Velocity: " + sVelocity).c_str(), 0, 48, 16, 16);
		disp.text(("Health: " + sHealth).c_str(), 0, 64, 16, 16);
	}
	if (displayMessage)
	{
		disp.text(message.c_str(), disp.width() / 2 - message.size() / 2 * 16, disp.height() / 2 - 8, 16, 16);
	}
	if (variableTable["fps"] && fps)
		disp.text(sfps.c_str(), disp.width() - 16 * sfps.size(), disp.height() - 16, 16, 16);
	if (splashScreen)
	{
		disp.select(texSplashScreen);
		disp.rectangle(0, disp.width(), 0, disp.height());
	}
	disp.select(font);
	if (variableTable["menu"] && activeMenu < menu.size())
		disp.display(menu[activeMenu]);
	if (consoleActive)
		disp.display(console);
	disp.depthTest(true);
}

void GameCore::resize(int width, int height)
{
	disp.resize(width, height);
}

void GameCore::processCommandList(const std::string &command)
{
	std::istringstream ss(command);
	std::string line;
	while (std::getline(ss, line, ';'))
		processCommand(line);
}

void GameCore::processCommand(const std::string &command)
{
	std::istringstream ss(command);
	std::string token;
	std::cout << "processCommand: \"" << command << "\"" << std::endl;
	ss >> token;
	if (token == "set")
	{
		int value = 0;
		ss >> token >> value;
		std::map<std::string, int>::iterator i = variableTable.find(token);
		if (i != variableTable.end())
		{
			i->second = value;
		}
		else
		{
			console.print("Invalid variable.");
		}
	}
	else if (token == "tanks")
	{
		int tankCount;
		ss >> tankCount;
		placeTanks(tankCount);
	}
	else if (token == "terrain")
	{
		splashScreen = false;
		terrain.generate(300, 0.25f, 0.75);
	}
	else if (token == "menu")
	{
		ss >> token;
		for (std::vector<MenuData>::iterator m = menu.begin(); m != menu.end(); m++)
		{
			if (m->id == token)
			{
				activeMenu = m - menu.begin();
				break;
			}
		}
		variableTable["menu"] = true;
	}
	else if (token == "work")
	{
		console.print("Error processing command.  Please bang head against wall.");
	}
	else if (token == "quit" || token == "exit")
		exit(0);
}

unsigned int GameCore::loadImage(const char *file)
{
	Image img;
	if (load(file, img))
		return disp.load(img);
	return 0;
}

void GameCore::placeTanks(int number)
{
	// First, we have to clear out all of the old tanks and objects...
	for (std::list<Object *>::iterator i = objects.begin(); i != objects.end(); i++)
		delete *i;
	
	objects.erase(objects.begin(), objects.end());
	tanks.erase(tanks.begin(), tanks.end());

	for (int t = 0; t < number; t++)
	{
		std::ostringstream ss;
		ss << "Player " << t + 1;
		TankObject *newTank = new TankObject;
		newTank->name(ss.str().c_str());
		newTank->skeleton = skeleton;
		newTank->texture = imageID;
		newTank->radius = 3.0f;
		newTank->position.x = 100.0f + 75.0f * sin(static_cast<float>(t) / number * 2 * 3.14159);
		newTank->position.z = 100.0f + 75.0f * cos(static_cast<float>(t) / number * 2 * 3.14159);
		newTank->position.y = 175.0f;
		tanks.push_back(newTank);
		objects.push_back(newTank);
	}

	tank = tanks.begin();
	getJoints();
}

void GameCore::getJoints()
{
	jointTable.erase(jointTable.begin(), jointTable.end());
	if (tank != tanks.end())
	{
		std::stack<SkeletonNode *> stack;
		stack.push(&(*tank)->skeleton);

		while (stack.size())
		{
			SkeletonNode *node = stack.top();
			stack.pop();

			for (std::vector<SkeletonNode>::iterator child = node->child.begin(); child != node->child.end(); stack.push(&*child++)) {}

			jointTable[node->id] = node;
		}
	}
}
