#include "menu.h"

#include "xmlparser.h"

#include <vector>
#include <string>
#include <iostream>

class MenuParser : public XMLParser
{
	std::vector<MenuData> &menu;
	std::vector<std::string> tagStack;
	MenuData currentMenu;
	MenuItem menuItem;
public:
	MenuParser(std::vector<MenuData> &newMenu) : menu(newMenu) {}
	void start(void *userData, const XML_Char *name, const XML_Char **atts);
	void end(void *userData, const XML_Char *name);
	void data(void *userData, const XML_Char *s, int len);
};

void MenuParser::start(void *userData, const XML_Char *name, const XML_Char **atts)
{
	tagStack.push_back(name);
	
	if (std::string(name) == "menulist")
		menu.resize(0);

	if (std::string(name) == "option") menuItem = MenuItem("","","");

	if (std::string(name) == "menu")
	{
		currentMenu.data.resize(0);
		currentMenu.selected = 0;
	}
}

void MenuParser::end(void *userData, const XML_Char *name)
{
	if (std::string(name) == "menu")
		menu.push_back(currentMenu);
	if (std::string(name) == "option")
		currentMenu.data.push_back(menuItem);

	tagStack.pop_back();
}

void MenuParser::data(void *userData, const XML_Char *s, int len)
{
	if (tagStack.size())
	{
		std::string tag = *tagStack.rbegin();
		if (tag == "id" && tagStack.size() > 2)
		{
			std::string previous = tagStack[tagStack.size() - 2];
			if (previous == "option")
				menuItem.id = std::string(s, s + len);
			else if (previous == "menu")
				currentMenu.id = std::string(s, s + len);
		}
		else if (tag == "label")
			menuItem.label = std::string(s, s + len);
		else if (tag == "execute")
			menuItem.execute = std::string(s, s + len);
		else if(tag == "choice" && tagStack.size() > 2)
			menuItem.choices.push_back(std::string(s, s + len));
	}
}

bool load(const char *filename, std::vector<MenuData> &menus)
{
	MenuParser parser(menus);
	return parser.parse(filename);
}
