#include "missiles.h"

ExplosiveWeapon::ExplosiveWeapon(Type t)
: explosiveTexture(0), explosionTexture(0), iconTexture(0)
{
	switch (t)
	{
	case BabyMissile:	
		radius = 4.0f;
		weaponName = "Baby Missile";
		weaponIcon = "data/weapons/babymissile.png";
		force = 50.0f;
		break;
	case Missile:
		radius = 8.0f;
		weaponName = "Missile";
		weaponIcon = "data/weapons/missile.png";
		force = 75.0f;
		break;
	case BabyNuke:
		radius = 16.0f;
		weaponName = "Baby Nuke";
		weaponIcon = "data/weapons/babynuke.png";
		force = 100.0f;
		break;
	case Nuke:
		radius = 32.0f;
		weaponName = "Nuke";
		weaponIcon = "data/weapons/nuke.png";
		force = 125.0f;
		break;
	}
}

bool ExplosiveWeapon::load(std::map<std::string, Mesh> &meshMap, Display &display)
{
	Image image;
	bool result = true;
	result &= ::load("data/sphere.skel", explosiveSkeleton);
	result &= ::load("data/explosion.skel", explosionSkeleton);
	result &= ::load("data/giraffe.png", image);
	explosiveTexture = display.load(image);
	result &= ::load("data/fire.png", image);
	explosionTexture = display.load(image);
	result &= ::load(weaponIcon, image);
	iconTexture = display.load(image);

	return result;
}

void ExplosiveWeapon::fire(std::list<Object *> &objects, const vertex &position, const vertex &velocity)
{
	Explosive *newExplosive = new Explosive;
	Explosion *newExplosion = new Explosion(radius, force);

	newExplosive->skeleton = explosiveSkeleton;
	newExplosive->texture = explosiveTexture;
	newExplosive->position = position;
	newExplosive->velocity = velocity;

	newExplosion->skeleton = explosionSkeleton;
	newExplosion->texture = explosionTexture;
	newExplosion->active = false;
	newExplosion->resting = true;
	newExplosion->velocity.x = newExplosion->velocity.y = newExplosion->velocity.z = 0.0f;

	newExplosive->explosion = newExplosion->id();

	objects.push_back(newExplosive);
	objects.push_back(newExplosion);
}
