#ifndef PHYSICSH
#define PHYSICSH

#include <list>

// #include "object.h"
#include "vertex.h"
#include "terrain.h"

class Object;

class Physics
{
public:
	struct Wall
	{
		enum Type { bounce };
		Wall(float nLeft, float nRight, float nBottom, float nTop, float nBack, float nFront, Type nType)
			: left(nLeft), right(nRight), bottom(nBottom), top(nTop), back(nBack), front(nFront), type (nType) {}
		float left;
		float right;
		float bottom;
		float top;
		float back;
		float front;
		Type type;
	};
	struct Environment
	{
		Environment(std::list<Object *> &newObjects, const Wall &newWall, Terrain &newTerrain) : objects(&newObjects), wall(&newWall), terrain(&newTerrain) {}
		std::list<Object *> *objects;
		const Wall *wall;
		Terrain *terrain;
	};
private:
	float timeStep;
	vertex gravity;
	
	float getNextCollision(std::list<Object *> &objects, const Wall &wall, Object *&a, vertex &bounceFactor);
	float getNextCollision(std::list<Object *> &objects, Object *&a, Object *&b);
	float getNextCollision(std::list<Object *> &objects, Terrain &terrain, Object *&a);
	void advanceObjects(Environment env, float timeToAdvance);
	void reactToCollision(Object &a, Object &b);
	void advanceState(Environment env /*std::list<Object> &objects, const Wall &wall, Terrain &terrain*/);

	bool allObjectsStationary(const std::list<Object *> &objects);

	float sphereSphereCollision(const Object &a, const Object &b, float t);
	float sphereSphereIntersection(const Object &a, const Object &b, float t);
public:
	Physics() { timeStep = 0.01f; gravity.x = 0.0f; gravity.y = -9.81f; gravity.z = 0.0f; }

	void determineFinalState(Environment env/*std::list<Object> &objects, const Wall &wall, Terrain &terrain*/);
	void advanceState(Environment env, float timeToAdvance /*std::list<Object> &objects, float timeToAdvance, const Wall &wall, Terrain &terrain*/);
};


#endif
