#include "tmatrix.h"

#include <cmath>

tMatrix::tMatrix()
{
	loadIdentity();
}

void tMatrix::loadIdentity()
{
	for (int row = 0; row < 4; row++)
	{
		for (int col = 0; col < 4; col++)
		{
			data[row][col] = (row==col)?1.0f:0.0f;
		}
	}
}

tMatrix tMatrix::operator*(const tMatrix &a)
{
	tMatrix b;
	
	for (int row = 0; row < 4; row++)
	{
		for (int col = 0; col < 4; col++)
		{
			b.data[row][col] = 0;
			
			for (int pos = 0; pos < 4; pos++)
			{
				b.data[row][col] += data[row][pos] * a.data[pos][col];
			}
		}
	}

	return b;
}

tMatrix tMatrix::operator+(const tMatrix &a)
{
	tMatrix b;

	for (int row = 0; row < 4; row++)
	{
		for (int col = 0; col < 4; col++)
		{
			b.data[row][col] = data[row][col] + a.data[row][col];
		}
	}

	return b;
}

vertex tMatrix::operator*(const vertex &a)
{
	vertex v;
	v.x = a.x * data[0][0] + a.y * data[0][1] + a.z * data[0][2] + data[0][3];
	v.y = a.x * data[1][0] + a.y * data[1][1] + a.z * data[1][2] + data[1][3];
	v.z = a.x * data[2][0] + a.y * data[2][1] + a.z * data[2][2] + data[2][3];
	return v;
}

void tMatrix::translate(const vertex &a)
{
	tMatrix T;
	T.data[0][3] = a.x;
	T.data[1][3] = a.y;
	T.data[2][3] = a.z;

	*this = *this * T;
}

void tMatrix::rotate(const vertex &a)
{
	float deg2rad = 3.14159f / 180.0f;
	float sinx = static_cast<float>(sin(a.x * deg2rad));
	float cosx = static_cast<float>(cos(a.x * deg2rad));
	float siny = static_cast<float>(sin(a.y * deg2rad));
	float cosy = static_cast<float>(cos(a.y * deg2rad));
	float sinz = static_cast<float>(sin(a.z * deg2rad));
	float cosz = static_cast<float>(cos(a.z * deg2rad));

	if (a.x)
	{
		tMatrix Rx;
		Rx.data[1][1] = cosx; Rx.data[1][2] = -sinx;
		Rx.data[2][1] = sinx; Rx.data[2][2] = cosx;
		*this = *this * Rx;
	}

	if (a.y)
	{
		tMatrix Ry;
		Ry.data[0][0] = cosy; Ry.data[0][2] = siny;
		Ry.data[2][0] = -siny; Ry.data[2][2] = cosy;
		*this = *this * Ry;
	}

	if (a.z)
	{
		tMatrix Rz;
		Rz.data[0][0] = cosz; Rz.data[0][1] = -sinz;
		Rz.data[1][0] = sinz; Rz.data[1][1] = cosz;
		*this = *this * Rz;
	}
}

void tMatrix::scale(const vertex &a)
{
	tMatrix S;

	S.data[0][0] = a.x;
	S.data[1][1] = a.y;
	S.data[2][2] = a.z;

	*this = *this * S;
}

std::ostream &operator<<(std::ostream &stream, const tMatrix &m)
{
	stream << '[';
	
	for (int row = 0; row < 4; row++)
	{
		stream << '[';
		for (int col = 0; col < 4; col++)
		{
			stream << '\t' << m.data[row][col];
		}
		stream << ']';
		if (row != 3) stream << std::endl;
	}

	stream << ']' << std::endl;

	return stream;
}
