#ifndef WINDOWH
#define WINDOWH

#include "gamecore.h"
//#include "sdl.h"

#include <map>

#ifdef __APPLE__
#include <Glut/glut.h>
#else
#include <GL/glut.h>
#endif

class Window
{
	int videoFlags;
	int colorDepth;
	int width;
	int height;

	int windowID;
	static Window *currentWindow;
	GameCore *game;

	bool leftDown;
	bool middleDown;
	bool rightDown;

	int lastX;
	int lastY;

	bool down;
public:
	Window(int &argc, char *argv[], const char *caption = "Untitled", int newWidth=640, int newHeight=480);
	~Window();

	void eventLoop(GameCore *newGame);

	static void staticDisplay() { if (currentWindow) currentWindow->display(); }
	static void staticIdle() { if (currentWindow) currentWindow->idle(); }
	static void staticReshape(int newWidth, int newHeight) { if (currentWindow) currentWindow->reshape(newWidth, newHeight); }
	static void staticKeyboard(unsigned char key, int x, int y) { if (currentWindow) currentWindow->keyboard(key, x, y, true); }
	static void staticKeyboardUp(unsigned char key, int x, int y) { if (currentWindow) currentWindow->keyboard(key, x, y, false); }
	static void staticSpecial(int key, int x, int y) { if (currentWindow) currentWindow->special(key, x, y, true); }
	static void staticSpecialUp(int key, int x, int y) { if (currentWindow) currentWindow->special(key, x, y, false); }
	static void staticMouse(int button, int state, int x, int y) { if (currentWindow) currentWindow->mouse(button, state, x, y); }
	static void staticMotion(int x, int y) { if (currentWindow) currentWindow->motion(x, y); }
	
	void display();
	void idle();
	void reshape(int width, int height);
	void keyboard(unsigned char key, int x, int y, bool down);
	void special(int key, int x, int y, bool down);
	void mouse(int button, int state, int x, int y);
	void motion(int x, int y);

	// static void staticPassiveMotion(int x, int y);
};

enum TankWindowKey
{
	TWK_BACKSPACE = 8,
	TWK_ENTER = 13,
	TWK_ESCAPE = 27,
	TWK_LEFT = 128,
	TWK_RIGHT = 129,
	TWK_UP = 130,
	TWK_DOWN = 131,
	TWK_PAGEUP = 132,
	TWK_PAGEDOWN = 133
};

#define TWK_SHIFT GLUT_ACTIVE_SHIFT
#define TWK_ALT GLUT_ACTIVE_ALT
#define TWK_CTRL GLUT_ACTIVE_CTRL

#endif
